/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)
  Coding Technologies

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996, 2007.
 *                                                                           *
 ****************************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "mps_conf.h"
#include "spatial_filereader.h"
#include "sac_bitinput.h"

int GetSpatialSpecificConfigData(char *file, int *pbsTreeConfig, int *pbsGainDMX,
                                 int *pbsFrameLength) {
  HANDLE_FILE_READER fileReader = NULL;
  HANDLE_BYTE_READER bytereader = NULL;
  HANDLE_S_BITINPUT bitstream;

  int sacTimeAlignFlag;
  int sacHeaderLen;

  int bsSamplingFreqIndex;
  int bsSamplingFrequency;
  int bsFrameLength;
  int bsFreqRes;
  int bsTreeConfig;
  int bsQuantMode;
  int bsOneIcc;
  int bsArbitraryDownmix;
  int bsFixedGainDMX;
  int bsFixedGainSur;
  int bsFixedGainLFE;

  fileReader = FileReaderOpen(file);
  bytereader = FileReaderGetByteReader(fileReader);
  bitstream = s_bitinput_open(bytereader);

  if (bitstream == 0) {
    return 0;
  }

  sacTimeAlignFlag = s_GetBits(bitstream, 1);
  sacHeaderLen = s_GetBits(bitstream, 7);
  if (sacHeaderLen == 127) {
    sacHeaderLen += s_GetBits(bitstream, 16);
  }

  bsSamplingFreqIndex = s_GetBits(bitstream, 4);
  if(bsSamplingFreqIndex == 15) {
    bsSamplingFrequency = s_GetBits(bitstream, 24);
  }
  bsFrameLength = s_GetBits(bitstream, 7);
  bsFreqRes = s_GetBits(bitstream, 3);
  bsTreeConfig = s_GetBits(bitstream, 4);
  bsQuantMode        = s_GetBits(bitstream, 2);
  bsOneIcc           = s_GetBits(bitstream, 1);
  bsArbitraryDownmix = s_GetBits(bitstream, 1);
  bsFixedGainSur     = s_GetBits(bitstream, 3);
  bsFixedGainLFE     = s_GetBits(bitstream, 3);
  bsFixedGainDMX     = s_GetBits(bitstream, 3);

  s_bitinput_close(bitstream);

  *pbsTreeConfig = bsTreeConfig;
  *pbsGainDMX = bsFixedGainDMX;
  *pbsFrameLength = bsFrameLength;

  return 1;
}

