/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)
  Coding Technologies

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996, 2007.
 *                                                                           *
 ****************************************************************************/

#include <math.h>
#include <string.h>

#include "audiofile_io.h"
#include "mps_conf.h"
#include "bitinput.h"
#include "ct_polyphase.h"
#include "adifdec.h"

#define USE_THRESHOLDS_NO

extern Twiddle *twiddleExtern;

static int
getThresholds(int upmixType,  int binauralQuality, char *inputfile,
              int *thrMaxQmf, int *thrMax, float *thrRmsQmf, float *thrQmf);

static int
getGroupIndex(char *inputfile);

int DiffPcm(char *file1, char *file2, int npart, int nSmpPerFrame,
            int *pmax1, double *prms1, int *pmax2, double *prms2) {
  AFILE *pcmStream1;
  AFILE *pcmStream2; 
  float pcmBuffer1[Chans*MSMP];
  float pcmBuffer2[Chans*MSMP];
  audioheader audio_header1,audio_header2;
  int samples_read1, samples_read2;
  int sample_counter;
  int nch, ismp, nframe, iframe;
  int ipart = 0;
  float diff;
  int count[2] = { 0, 0 };
  float maxDiff[2]   = { 0.0f, 0.0f };
  float totalDiff[2] = { 0.0f, 0.0f };

  if (open_inputaudio(&pcmStream1,
                      file1,
                      &audio_header1)) {
    return 1;
  }

  if (open_inputaudio(&pcmStream2,
                      file2,
                      &audio_header2)) {
    return 1;
  }
      
  nframe = audio_header1.nrofsamples/nSmpPerFrame;
  nch = audio_header1.channels;

  sample_counter = 0;
  for (iframe = 0; iframe < nframe; iframe++) {
    if (iframe == nframe/npart) {
      ipart = 1;
    }

    samples_read1 = get_audiosample(pcmBuffer1,
                                    pcmStream1,
                                    nch,
                                    sample_counter,
                                    nSmpPerFrame);

    samples_read2 = get_audiosample(pcmBuffer2,
                                    pcmStream2,
                                    nch,
                                    sample_counter,
                                    nSmpPerFrame);

    if (samples_read1 != samples_read2) {
      break;
    }

    sample_counter += samples_read1;

    for (ismp = 0; ismp < nSmpPerFrame*nch; ismp++) {
      count[ipart]++;
      diff = (float)fabs(pcmBuffer1[ismp] - pcmBuffer2[ismp]);
      if (diff > maxDiff[ipart] && iframe) {
        maxDiff[ipart] = diff;
      }

      totalDiff[ipart] += diff*diff;
    }
  }

  AFclose(pcmStream1);
  AFclose(pcmStream2);

  if (npart == 1) {
    if (count[0] == 0) {
      return 1;
    }
  }
  else if (npart == 2) {
    if (count[0] == 0 || count[1] == 0) {
      return 1;
    }
  }

  *pmax1 = (int)(maxDiff[0] + 0.5f);
  *prms1 = sqrt(totalDiff[0]/count[0]);

  if (npart == 2) {
    *pmax2 = (int)(maxDiff[1] + 0.5f);
    *prms2 = sqrt(totalDiff[1]/count[1]);
  }

  return 0;
}


void ShowConformanceResult(int isBuriedData, int isSbr, int upmixType,
                           int binauralQuality, char *inputfile,
                           int npart, int max1, double rms1, int max2, double rms2) {
  int isQmfTest, ierr;
  int thrMaxQmf, thrMax;
  float thrRmsQmf, thrRms;
  int failed = 0;

#ifdef PARTIALLY_COMPLEX
  PRINT(SO,"\n\nLow Power ");
#else
  PRINT(SO, "\n\nHigh Quality ");
#endif

  PRINT(SO, "MPS conformance test on %s \n\n", inputfile);

  isQmfTest = 1;
  if (isBuriedData || upmixType == 1) {
    isQmfTest = 0;
  }

  switch (upmixType) {
  case 0:
    PRINT(SO, "Multichannel output\n");
    break;
  case 1:
    PRINT(SO, "Enhanced matrix mode\n");
    isQmfTest = 0;
    break;
  case 2:
    if (binauralQuality == 0) {
      PRINT(SO, "Binaural output low power\n");
    }
    else if (binauralQuality == 1) {
      PRINT(SO, "Binaural output high quality\n");
    }

    break;
  case 3:
    PRINT(SO, "Stereo output\n");
    break;
  default:
    return;
  }

#ifndef PARTIALLY_COMPLEX
  if (isSbr) {
    if (twiddleExtern != NULL) {
      PRINT(SO, "Twiddle fi = %.3f, beta = %.3f\n", twiddleExtern->fi, twiddleExtern->beta);
    }
    else {
      PRINT(SO, "Default twiddle fi = 0.500, beta = 64.000\n");
    }
  }
#endif

  PRINT(SO, "\n");

  ierr = getThresholds(upmixType, binauralQuality, inputfile,
                       &thrMaxQmf, &thrMax, &thrRmsQmf, &thrRms);

  if (isQmfTest) {
    PRINT(SO, "QMF-TEST (Informative)\n");

    if (ierr == 0) {
      PRINT(SO, "  Max diff: %8d (%8d)\n", max1, thrMaxQmf);
      PRINT(SO, "  RMS diff: %8.3f (%8.3f)\n",  rms1, thrRmsQmf);
    }
    else {
      PRINT(SO, "  Max diff: %8d\n", max1);
      PRINT(SO, "  RMS diff: %8.3f\n",  rms1);
    }
  }
  else {
    PRINT(SO, "MPS-TEST\n");
    if (ierr == 0) {
      if (max1 > thrMax || rms1 > thrRms) {
        failed = 1;
      }
    }

    if (ierr == 0) {
      PRINT(SO, "  Max diff: %8d (%8d)\n", max1, thrMax);
      PRINT(SO, "  RMS diff: %8.3f (%8.3f)\n",  rms1, thrRms);
    }
    else {
      PRINT(SO, "  Max diff: %8d\n", max1);
      PRINT(SO, "  RMS diff: %8.3f\n",  rms1);
    }
  }

  if (npart == 2) {
    PRINT(SO, "\nMPS-TEST\n");
    if (ierr == 0) {
      if (max2 > thrMax || rms2 > thrRms) {
        failed = 1;
      }
    }

    if (ierr == 0) {
      PRINT(SO, "  Max diff: %8d (%8d)\n", max2, thrMax);
      PRINT(SO, "  RMS diff: %8.3f (%8.3f)\n",  rms2, thrRms);
    }
    else {
      PRINT(SO, "  Max diff: %8d\n", max2);
      PRINT(SO, "  RMS diff: %8.3f\n",  rms2);
    }
  }


  if (ierr == 0) {
    if (failed) {
      PRINT(SO,  "\nConformance FAILED\n");
    }
    else {
      PRINT(SO, "\nConformance PASSED\n");
    }
  }
  else   if (ierr == -1) {
#ifdef USE_THRESHOLDS
    PRINT(SO, "\nNo thresholds\n");
#endif
  }
  else   if (ierr == -2) {
    PRINT(SO, "\nFile name of bitstream not recognized\n");
  }
}

/* 
  column

  0 - upmixType = 0, multichannel
  1 - upmixType = 1, blind upmix
  2 - upmixType = 2, binaural output, single slot
  3 - upmixType = 3, stereo output
  4 - upmixType = 2, binaural output,multi slot
*/

#ifdef PARTIALLY_COMPLEX
const int critMax[][5] = {
  {40, 40, 40, 40, 100},
  {40, 40, 40, 40, 100},
  {40, 40, 40, 40, 100}
};

const float critRms[][5] = {
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f},
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f},
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f}
};
#else
const int critMax[][5] = {
  {40, 40, 40, 40, 100},
  {40, 40, 40, 40, 100},
  {40, 40, 40, 40, 100}
};

const float critRms[][5] = {
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f},
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f},
  {2.0f,  2.0f,  2.0f,  2.0f,  4.0f}
};
#endif

static int
getThresholds(int upmixType, int binauralQuality, char *inputfile,
              int *pthrMaxQmf, int *pthrMax, float *pthrRmsQmf, float *pthrRms) {
  int icrit;
  int igroup = getGroupIndex(inputfile);

#ifndef USE_THRESHOLDS
  return -1;
#endif

  if (upmixType == 1) {
    igroup = 0;   /* blind upmix, pcm input file, don't care about file name */
  }

  if (upmixType == 2 && binauralQuality == 1) {
    icrit = 4;
  }
  else {
    icrit = upmixType;
  }

  *pthrMaxQmf = critMax[0][icrit];
  *pthrRmsQmf = critRms[0][icrit];

  if (igroup < 0) { /* File not in conformance file list */
    *pthrMax    = -1;
    *pthrRms    = -1.0f;

    return -2;
  }
  else {
    *pthrMax    = critMax[igroup][icrit];
    *pthrRms    = critRms[igroup][icrit];

    if (*pthrMax < 0) { /* No threshold */
      return -1;
    }
    else {
      return 0;
    }
  }
}

static const char *fileNames[] = {
  "heaac_mps_oneicc",
  "heaac_mps_param",
  "heaac_mps_qmf",
  "heaac_mps_quant",
  "heaac_mps_res",
  "heaac_mps_shape",
  "heaac_mps_smooth",
  "heaac_mps_tree",
  "heaac_mps_",

  "aac_mps_oneicc",
  "aac_mps_param",
  "aac_mps_qmf",
  "aac_mps_quant",
  "aac_mps_res",
  "aac_mps_shape",
  "aac_mps_smooth",
  "aac_mps_tree",
  "aac_mps_",

  "pcm_mps_32",
  "pcm_mps_44",
  "pcm_mps_arb",
  "pcm_mps_oneicc",
  "pcm_mps_param",
  "pcm_mps_quant",
  "pcm_mps_res",
  "pcm_mps_qmf",
  "pcm_mps_shape",
  "pcm_mps_smooth",
  "pcm_mps_tree",
  "pcm_mps_ts",
  "pcm_mps_",

  "l2_mps_32",
  "l2_mps_44",
  "l2_mps_arb",
  "l2_mps_mtx",
  "l2_mps_oneicc",
  "l2_mps_param",
  "l2_mps_quant",
  "l2_mps_res",
  "l2_mps_shape",
  "l2_mps_smooth",
  "l2_mps_tree",
  "l2_mps_ts",
  "l2_mps_",

  "l3_mps_32",
  "l3_mps_44",
  "l3_mps_arb",
  "l3_mps_mtx",
  "l3_mps_oneicc",
  "l3_mps_param",
  "l3_mps_quant",
  "l3_mps_res",
  "l3_mps_shape",
  "l3_mps_smooth",
  "l3_mps_tree",
  "l3_mps_ts",
  "l3_mps_",

  "end"
};

static const int fileToGroup[] = {
  2, 2, 0, 2, 2, 2, 2, 2, 2, /* HE-AAC */
  1, 1, 0, 1, 1, 1, 1, 1, 1, /* AAC */
  1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1,  /* buried data */
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,  /* L2 */
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1   /* L3 */
};

static int
getGroupIndex(char *name) {
  int iname = 0;

  while (strcmp(fileNames[iname], "end")) {
    if (strstr(name, fileNames[iname])) {
      return fileToGroup[iname];
    }

    iname++;
  }

  return -1;
}

