/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)
  Coding Technologies

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996, 2007.
 *                                                                           *
 ****************************************************************************/

#include <setjmp.h>
#include <string.h>
#include "sac_dec.h"
#include "libtsp.h"
#include <libtsp/AFpar.h>
#include "sac_polyphase.h"
#include "spatial_filereader.h"
#include "sac_parallelReverb.h"
#include "mps_conf.h"
#include "spatial_bdreader.h"

#define PRINT fprintf
#define SE stderr
#define MAX_QMF_BANDS 128 

typedef enum {
  BS_FILE,
  BS_BURIED,
  BS_NONE
} BITSTREAM_SOURCE;

#define DMX_GAIN_FIX

#ifdef HRTF_DYNAMIC_UPDATE
HANDLE_HRTF_SOURCE hrtfSource;
#endif

extern int upmixType;
extern int hrtfModel;
extern int binauralQuality;
extern int decType;

int debug[256];

jmp_buf g_JmpBuf;

/* MPEG Surround Decoder */

int
MpsDecode(int isBuriedData, char* pcmInFile, char* spatialBitstreamFile, char* spatialPcmOutFile)
{
  BITSTREAM_SOURCE bitstreamSource;
  spatialDec* ourDec;
  long int nSamples;
  long int nChannels;
  double sampleFreq;
  int qmfBands = 0;
  int nTimeSlots;

  AFILE* pcmIn = 0;
  HANDLE_FILE_READER fileReader = NULL;
  HANDLE_BURIED_DATA_READER bdReader = NULL;
  BD_INPUT_BUFFER bdBuffer;
  HANDLE_BYTE_READER bitstream = NULL;

  int done = 0;
  int samplesToReadPerCall;
  int samplesRead;
  int frameCounter = 0;
  float* inSamples;
  float* inSamplesDeinterleaved;
  int i, j, channel, ts;
  long int offset=0;

  SAC_POLYPHASE_ANA_FILTERBANK *filterbank[MAX_INPUT_CHANNELS] = {0};
  float** qmfInReal[MAX_INPUT_CHANNELS];
  float** qmfInImag[MAX_INPUT_CHANNELS];
  float** inPointers[2*MAX_INPUT_CHANNELS];

  if (isBuriedData) {
    bitstreamSource = BS_BURIED;
  }
  else {
    bitstreamSource = BS_FILE;
  }

  /* allocate qmf memory */
#ifdef PARTIALLY_COMPLEX
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    qmfInReal[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    for (j=0 ; j <  MAX_TIME_SLOTS; j++) {
      qmfInReal[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
    }
    inPointers[i] = qmfInReal[i];
    (qmfInImag);
  }
#else
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    qmfInReal[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    qmfInImag[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    for (j=0 ; j < MAX_TIME_SLOTS; j++) {
      qmfInReal[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
      qmfInImag[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
    }
    inPointers[2*i  ] = qmfInReal[i];
    inPointers[2*i+1] = qmfInImag[i];
  }
#endif

  if(setjmp(g_JmpBuf)==0) {
  }
  else {
    exit(-1);
  }

  if (pcmInFile)
    pcmIn = AFopnRead(pcmInFile,
                      &nSamples,
                      &nChannels,
                      &sampleFreq,
                      0);

  if (bitstreamSource == BS_FILE && upmixType != 1) {
    fileReader = FileReaderOpen(spatialBitstreamFile);
    bitstream = FileReaderGetByteReader(fileReader);
  }

  if (sampleFreq < 27713.0) {
    qmfBands = 32;
  } else if (sampleFreq >= 55426.0) {
    qmfBands = 128;
  } else {
    qmfBands = 64;
  }

  if (bitstreamSource == BS_BURIED) {
    float header[BD_HEADER_LENGTH * MAX_INPUT_CHANNELS];
    int headerSamples = BD_HEADER_LENGTH * nChannels;
    int samplesToRead;

    bdReader = BdReaderOpen(qmfBands);
    bitstream = BdReaderGetByteReader(bdReader);

    bdBuffer.channels      = nChannels;
    bdBuffer.sampleFormat  = pcmIn->Format;
    bdBuffer.bitsPerSample = pcmIn->NbS;
    bdBuffer.data          = header;

    offset += (samplesRead = AFfReadData(pcmIn, offset, header, headerSamples));
    if (samplesRead != headerSamples) done = 1;

    if (BdReaderGetFrameSize(bdReader, &bdBuffer, &nTimeSlots) != 0) done = 1;

    samplesToReadPerCall = nChannels * qmfBands * nTimeSlots;
    inSamples = (float*)calloc(sizeof(float), samplesToReadPerCall);

    bdBuffer.data = inSamples;

    memcpy(inSamples, header, sizeof(float) * headerSamples);
    samplesToRead = samplesToReadPerCall - headerSamples;

    offset += (samplesRead = AFfReadData(pcmIn, offset, &inSamples[headerSamples], samplesToRead));
    if (samplesRead == samplesToRead) {
      if (BdReaderParseFrame(bdReader, &bdBuffer) != 0) done = 1;
    }
    else done = 1;
  }

  ourDec = SpatialDecOpen(bitstream,
                          spatialPcmOutFile,
                          (int)sampleFreq,
                          (int)nChannels,
                          &nTimeSlots,
                          qmfBands,
                          decType,
                          upmixType,
                          hrtfModel,
#ifdef HRTF_DYNAMIC_UPDATE
                          hrtfSource,
#endif
                          binauralQuality);


  SacInitAnaFilterbank(NULL, qmfBands);
  for (channel = 0; channel < nChannels; channel++)
    SacOpenAnaFilterbank(&filterbank[channel]);

  inSamplesDeinterleaved = (float*) calloc(sizeof(float), qmfBands);

  if (bitstreamSource != BS_BURIED) {
    samplesToReadPerCall = nChannels * qmfBands * nTimeSlots;
    inSamples = (float*)calloc(sizeof(float), samplesToReadPerCall);
  }

  /* main loop */
  while (!done) {
    if ((bitstreamSource != BS_BURIED) || (frameCounter > 0)) {
      offset += (samplesRead = AFfReadData(pcmIn, offset, inSamples, samplesToReadPerCall));
      if (samplesRead == samplesToReadPerCall) {
        if (bitstreamSource == BS_BURIED) {
          if (BdReaderParseFrame(bdReader, &bdBuffer) != 0) done = 1;
          if (BdReaderEndOfStream(bdReader)) done = 1;
        }
      }
      else done = 1;
    }

    fprintf(stderr, "\rframe [%4d]", ++frameCounter);

    if (upmixType != 1) {
      if ((bitstreamSource == BS_FILE) && FileReaderEof(fileReader)) done = 1;
    }

    if ((!done) && (setjmp(g_JmpBuf)==0)) {

      if (upmixType != 1) {
        SpatialDecParseFrame(ourDec);
      }

#ifdef DMX_GAIN_FIX
      BinauralDoParallelReverb(ourDec, inSamples, 32768.0f);
#else
      BinauralDoParallelReverb(ourDec, inSamples, 32768.0f * SpatialGetClipProtectGain(ourDec));
#endif

      /* QMF analysis */
      for (channel=0; channel < nChannels; channel++) {
        /* deinterleave and scale */
        for (ts=0; ts< nTimeSlots; ts++) {
          for (i=0; i< qmfBands; i++) {
#ifdef DMX_GAIN_FIX
            if (upmixType == 2 || upmixType == 3)
              inSamplesDeinterleaved[i] = inSamples[nChannels* (ts*qmfBands+i)+channel] * 32768.0f;
            else
#endif
              inSamplesDeinterleaved[i] = inSamples[nChannels* (ts*qmfBands+i)+channel] * 32768.0f * SpatialGetClipProtectGain(ourDec);
          }

#ifdef PARTIALLY_COMPLEX
          SacCalculateAnaFilterbank( filterbank[channel],
                                     inSamplesDeinterleaved,
                                     inPointers[channel][ts],
                                     NULL );
#else
          SacCalculateAnaFilterbank( filterbank[channel],
                                     inSamplesDeinterleaved,
                                     inPointers[2*channel][ts],
                                     inPointers[2*channel+1][ts] );
#endif
        }
      }

      SpatialDecApplyFrame(ourDec,
                           nTimeSlots,
                           inPointers);

    }
    else {
      break;
    }
  }

  /* free input buffer */
  if (inSamples != NULL)             free(inSamples);
  if (inSamplesDeinterleaved != NULL) free(inSamplesDeinterleaved);

  for (channel = 0; channel < nChannels; channel++)
    SacCloseAnaFilterbank(filterbank[channel]);

  SpatialDecClose(ourDec);
  if (bitstreamSource == BS_BURIED) BdReaderClose(bdReader);
  if (bitstreamSource == BS_FILE) FileReaderClose(fileReader);
  AFclose (pcmIn);

  /* free qmf memory */
#ifdef PARTIALLY_COMPLEX
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    for (j=0 ; j < MAX_TIME_SLOTS ; j++) {
      if (qmfInReal[i][j] != NULL) free(qmfInReal[i][j]);
    }
    if (qmfInReal[i] != NULL) free(qmfInReal[i]);
  }
#else
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    for (j=0 ; j < MAX_TIME_SLOTS ; j++) {
      if (qmfInReal[i][j] != NULL) free(qmfInReal[i][j]);
      if (qmfInImag[i][j] != NULL) free(qmfInImag[i][j]);
    }
    if (qmfInReal[i] != NULL) free(qmfInReal[i]);
    if (qmfInImag[i] != NULL) free(qmfInImag[i]);
  }
#endif
  return 0;
}


static void
myexit(char *s)
{

  if (s) {
    PRINT(SE,"\nMYEXIT: %s\n", s);
  }
  longjmp(g_JmpBuf,1);
}

