/**********************************************************************
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 *
 * $Id: decoder.h,v 1.1.1.1 2008/02/29 17:08:19 purnhaho Exp $
 *
 * $Log: decoder.h,v $
 * Revision 1.1.1.1  2008/02/29 17:08:19  purnhaho
 * initial import of mp12_mps_tool.zip from Philips 2008-02-29
 *
 * Revision 1.14  2006/03/10 15:11:56  dodenhen
 * Added MPEG_2_5 support
 *
 * Revision 1.13  2004/10/04 14:08:45  kaehleof
 * added switch for alias reduction in mixed blocks
 *
 * Revision 1.12  2004/08/10 14:34:15  sps
 * added some code to allow for investigation of channel configuration
 * changes, to be enabled using the define CHANNEL_INVESTIGATION in
 * decoder.h
 *
 * Revision 1.11  2003/11/13 12:20:31  tmm01
 *   by Ernst F. Schroeder, look for: efs
 * + added routines for CRC check in Layer III
 *   CRC errors are displayed and counted -
 *   no further action is taken
 *
 * Revision 1.10  2003/05/13 17:15:34  sps
 *  + bug fix in SubBandSynthesis() (previously introduced while changing
 *    to automatic arrays (was previously malloc))
 *
 *  + revision of the out_fifo interface (only the error code is returned,
 *    the clipping information became a function parameter)
 *
 *  + revised the code such that it can be compiled using g++
 *
 *  + functions returning an error code have now the return type
 *    ERROR_CODE. This concerns
 *     - out_fifo()
 *     - SubBandSynthesis()
 *     - III_get_side_info()
 *     - III_huffman_decode()
 *     - read_syn_window()
 *     - open_bit_stream_r()
 *     - initialize_huffman()
 *     - GetArguments()
 *     - aiff_read_headers()
 *     - aiff_seek_to_sound()
 *     - aiff_write_headers()
 *
 * Revision 1.9  2003/05/13 13:53:29  sps
 *  + bug fix in SubBandSynthesis() (previously introduced while changing
 *    to automatic arrays (was previously malloc)
 *
 *  + revision of the interface to out_fifo (only the error code is
 *    returned, the clipping information became a function parameter)
 *
 * Revision 1.8  2003/05/09 17:41:48  sps
 * +++ removed memory leaks.
 * +++ revised termination behaviour for unusual cases (file could not be
 * opened, no data could be written into file, memory allocation failed). A
 * couple of functions now return an error code, that leads to a clean
 * termination in the top level function.
 * +++ simplified legacy write functionality.
 * +++ removed some unnecessary "extern" declaration in function
 * declarations in header files
 *
 * Revision 1.7  2003/04/30 17:39:53  sps
 * +++ fixed aiff header for native output +++ fixed rounding for 16 bit
 * native output (by this, separated rounding in individual function
 * castDoubleToLong(), which is now called somewhat later (not anymore in
 * SubBandSynthesis(), but in out_fifo())) +++ made sure that the last
 * samples are written into the output file +++ made sure that decoding
 * starts only if file opening was successfull +++ added AFsp library
 * support, that allows the output of WAVE files. The output of raw and
 * aiff files is still supported, but is in this case also handled by the
 * AFsp library +++ Note: If the AFsp lib is used, any particular support
 * of erroneous signals is disabled +++
 *
 * Revision 1.6  2003/03/11 10:07:19  sps
 * committed contribution from Mike Coleman to remove certain kinds of
 * compiler warnings
 *
 * Revision 1.5  2002/11/14 13:04:08  tmm01
 * warnings removed (Microsoft Visual Studio V.6.0)  23.10.2 W.Voessing
 *
 * Revision 1.2  1996/03/28 03:13:37  rowlands
 * Merged layers 1-2 and layer 3 revisions
 *
 * Revision 1.1  1996/02/14 03:45:52  rowlands
 * Initial revision
 *
 * Received from FhG
 **********************************************************************/
/**********************************************************************
 *   date   programmers         comment                               *
 * 2/25/91  Doulas Wong,        start of version 1.0 records          *
 *          Davis Pan                                                 *
 * 5/10/91  Vish (PRISM)        Renamed and regrouped all ".h" files  *
 *                              into "common.h" and "decoder.h".      *
 *                              Ported to Macintosh and Unix.         *
 * 27jun91  dpwe (Aware)        New prototype for out_fifo()          *
 *                              Moved "alloc_*" stuff to common.h     *
 *                              Use ifdef PROTO_ARGS for prototypes   *
 *                              prototypes reflect frame_params struct*
 * 10/3/91  Don H. Lee          implemented CRC-16 error protection   *
 * 2/11/92  W. Joseph Carter    Ported new code to Macintosh.  Most   *
 *                              important fixes involved changing     *
 *                              16-bit ints to long or unsigned in    *
 *                              bit alloc routines for quant of 65535 *
 *                              and passing proper function args.     *
 *                              Removed "Other Joint Stereo" option   *
 *                              and made bitrate be total channel     *
 *                              bitrate, irrespective of the mode.    *
 *                              Fixed many small bugs & reorganized.  *
 *                              Modified some function prototypes.    *
 * 08/07/92 Mike Coleman        Made small changes for portability    *
 * 9/07/93  Toshiyuki Ishino    Integrated with Layer III.            *
 * 11/04/94 Jon Rowlands        fix protos for usage() and            *
 *                              recover_CRC_error()                   *
 *          Roland Bitto 	Adapted to MPEG2 low bitrate        	      *
 *--------------------------------------------------------------------*
 * 05/08/02 Walter Voessing     new functions for MPEG1 and MPEG 2    *
 *                              Stereo Processing                     *  
 *                              bugs within 'III_get_scale_factors'   *   
 *                              fixed                                 *
 *                              'out_fifo': bugs fixed and 24 bit     *
 *                                          output added	      *
 *                              'GetArguments': '-C' parameter added  *
 *                              (optional 24 bit output)              *
 *                              modified usage()                      *
 **********************************************************************/

/***********************************************************************
*
*  Decoder Include Files
*
***********************************************************************/

/***********************************************************************
*
*  Decoder Definitions
*
***********************************************************************/

#ifdef USELIB_AFsp
#include <libtsp.h>
#endif

#include "common.h"

#define   DFLT_OPEXT        ".dec"  /* default output file name extension */
#define   DFLT_MPSEXT       ".mps"  /* default MPEG Surround data file name extension */

/*
  this define enables some debug output to investigate changes in the 
  channel configuration

#define CHANNEL_INVESTIGATION
*/

/*
 NOTE: The value of a multiple-character constant is
 implementation-defined.
*/
#if !defined(MS_DOS) && !defined(AIX)
#define   FILTYP_DEC_AIFF   'AIFF'
#define   FILTYP_DEC_BNRY   'TEXT'
#define   CREATR_DEC_AIFF   'Sd2a'
/*
  The following character constant is ASCII '????'
  It is declared in hex because the character
  constant contains a trigraph, causing an error in
  parsing with ANSI preprocessors.
*/
#define   CREATR_DEC_BNRY   0x3f3f3f3f
#else
#define   FILTYP_DEC_AIFF   "AIFF"
#define   FILTYP_DEC_BNRY   "TEXT"
#define   CREATR_DEC_AIFF   "Sd2a"
#define   CREATR_DEC_BNRY   "????"
#endif


#ifdef MPEG_2_5 
#define	  SYNC_WORD	    (long) 0x7ff
#define	  SYNC_WORD_LNGTH   11
#else
#define   SYNC_WORD         (long) 0xfff 
#define   SYNC_WORD_LNGTH   12
#endif 

#define   MUTE              0

/***********************************************************************
*
*  Decoder Type Definitions
*
***********************************************************************/

/***********************************************************************
*
*  Decoder Variable External Declarations
*
***********************************************************************/

/***********************************************************************
*
*  Decoder Function Prototype Declarations
*
***********************************************************************/

/* The following functions are in the file "musicout.c" */


/* The following functions are in the file "decode.c" */

#ifdef   PROTO_ARGS
void   decode_info(Bit_stream_struc*, frame_params*);
void   II_decode_bitalloc(Bit_stream_struc*, unsigned int[2][SBLIMIT],
                       frame_params*);
void   I_decode_bitalloc(Bit_stream_struc*, unsigned int[2][SBLIMIT],
                       frame_params*);
void   I_decode_scale(Bit_stream_struc*, unsigned int[2][SBLIMIT],
                       unsigned int[2][3][SBLIMIT], frame_params*);
void   II_decode_scale(Bit_stream_struc*, unsigned int[2][SBLIMIT],
                       unsigned int[2][SBLIMIT], unsigned int[2][3][SBLIMIT],
                       frame_params*);
void   I_buffer_sample(Bit_stream_struc*, unsigned int[2][3][SBLIMIT],
                       unsigned int[2][SBLIMIT], frame_params*);
void   II_buffer_sample(Bit_stream_struc*, unsigned int[2][3][SBLIMIT],
                       unsigned int[2][SBLIMIT], frame_params*);
void   read_quantizer_table(double[17], double[17]);
void   II_dequantize_sample(unsigned int[2][3][SBLIMIT], 
                       unsigned int[2][SBLIMIT], double[2][3][SBLIMIT],
                       frame_params*);
void   I_dequantize_sample(unsigned int[2][3][SBLIMIT],
                       double[2][3][SBLIMIT], unsigned int[2][SBLIMIT],
                       frame_params*);
void   read_scale_factor(double[SCALE_RANGE]);
void   II_denormalize_sample(double[2][3][SBLIMIT],
                       unsigned int[2][3][SBLIMIT], frame_params*, int);
void   I_denormalize_sample(double[2][3][SBLIMIT],
                       unsigned int[2][3][SBLIMIT], frame_params*);
void   create_syn_filter(double[64][SBLIMIT]);

ERROR_CODE SubBandSynthesis (double*, 
                             int, 
                             double*
                             );

ERROR_CODE read_syn_window(double[HAN_SIZE]);
void   window_sample(double*, double*);

ERROR_CODE   out_fifo(
                      double[2][SSLIMIT][SBLIMIT], 
                      int, frame_params*, int,
#ifdef USELIB_AFsp
                      AFILE*,
#else
                      FILE*, 
#endif
                      unsigned long*, 
                      unsigned char, 
                      unsigned short 
                      *clip);

void   buffer_CRC(Bit_stream_struc*, unsigned int*);
#ifndef USELIB_AFsp
void   recover_CRC_error(double[2][SSLIMIT][SBLIMIT], int, frame_params*,
                         FILE*, unsigned long*, unsigned char);
#endif
void   III_dequantize_sample(long int[SBLIMIT][SSLIMIT],
                             double [SBLIMIT][SSLIMIT], III_scalefac_t *,
                             struct gr_info_s *, int, frame_params *);
void   III_antialias(double[SBLIMIT][SSLIMIT], double[SBLIMIT][SSLIMIT], 
                     struct gr_info_s *,
                     int antialias_mixed_blocks);
void   inv_mdct(double[18], double[36], int);
void   III_hybrid(double[SSLIMIT], double[SSLIMIT] , int, int,
                       struct gr_info_s *);
ERROR_CODE III_get_side_info_crc(Bit_stream_struc *,
                             III_side_info_t *,
                             frame_params *,
 							 unsigned int *crc);  /* added by efs 2003-11-11 */

void   III_put_side_info(Bit_stream_struc *,
                         III_side_info_t *,
                         frame_params *);


void III_get_scale_factors(III_scalefac_t *,
                           III_side_info_t *, 
                           int, 
                           int);

void III_get_LSF_scale_factors(III_scalefac_t *scalefac, III_side_info_t *si, int gr, int ch, frame_params *fr_ps);
ERROR_CODE III_huffman_decode(long int is[SBLIMIT][SSLIMIT], III_side_info_t *si, int ch, int gr, int part2_start, frame_params *fr_ps);
void III_reorder (double xr[SBLIMIT][SSLIMIT], double ro[SBLIMIT][SSLIMIT], struct gr_info_s *gr_info, frame_params *fr_ps); 

int main_data_slots( frame_params fr_ps );

/*********************************************************/
/* NEW FUNCTIONS FOR MPEG I AND MPEG 2 STEREO PROCESSING */
/*********************************************************/
/* WV, 24.10.02 */

void MPEG_1_III_stereo(double xr[2][SBLIMIT][SSLIMIT], double lr[2][SBLIMIT][SSLIMIT], III_scalefac_t *scalefac,
			        								struct gr_info_s *gr_info, frame_params *fr_ps);
void MPEG_2_III_stereo(double xr[2][SBLIMIT][SSLIMIT], double lr[2][SBLIMIT][SSLIMIT], III_scalefac_t *scalefac,
											        struct gr_info_s *gr_info_l, struct gr_info_s *gr_info_r, frame_params *fr_ps);


#else
void   decode_info();
void   II_decode_bitalloc();
void   I_decode_bitalloc();
void   I_decode_scale();
void   II_decode_scale();
void   I_buffer_sample();
void   II_buffer_sample();
void   read_quantizer_table();
void   II_dequantize_sample();
void   I_dequantize_sample();
void   read_scale_factor();
void   II_denormalize_sample();
void   I_denormalize_sample();
void   create_syn_filter();
int    SubBandSynthesis ();
ERROR_CODE read_syn_window();
void   window_sample();
unsigned short   out_fifo();
void   buffer_CRC();
void   recover_CRC_error();
void   III_dequantize_sample();
void   III_antialias();
void   inv_mdct();
void   III_hybrid();
void III_get_LSF_scale_factors();
ERROR_CODE  III_huffman_decode();
void III_reorder (); 
void MPEG_1_III_stereo();
void MPEG_2_III_stereo();
#endif
